"""
This blocking operation will call openai with instructor. It expects 
a pydantic model and a list of messages, and will return a BlockingOperation
which will call the next_function_to_call of the node's action with the completion
and context.
"""
from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, Dict, List

from pydantic import BaseModel

from swarmstar.swarm.types import BlockingOperation, SwarmState
from swarmstar.utils.ai.openai_instructor import completion

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


class expected_args(BaseModel):
    messages: List[
        Dict[str, str]
    ]  # This should be a list of dictionaries with the keys 'role' and 'content'
    instructor_model_name: str  # This should point to a pydnatic model in the swarmstar.utils.ai.openai_instructor.models module


def blocking(
    swarm: SwarmConfig, blocking_operation: BlockingOperation
) -> BlockingOperation:
    messages = blocking_operation.args["messages"]
    instructor_model_name = blocking_operation.args["instructor_model_name"]
    swarm_state = SwarmState(swarm=swarm)
    node = swarm_state[blocking_operation.node_id]
    
    
    
    models_module = import_module(
        "swarmstar.utils.swarm_utils.blocking_operations.instructor.pydantic_models"
    )
    instructor_model = getattr(models_module, instructor_model_name)

    response = completion(
        messages=messages,
        openai_key=swarm.openai_key,
        instructor_model=instructor_model,
    )

    node.developer_logs.append({
        "type": "instructor_request",
        "messages": messages
    })
    node.developer_logs.append({
        "type": "instructor_completion",
        "instructor_model_name": instructor_model_name,
        "completion": completion.model_dump_json(indent=4)
    })
    swarm_state.update_state(node)
    
    return BlockingOperation(
        node_id=blocking_operation.node_id,
        blocking_type="internal_action",
        args={"completion": response},
        context=blocking_operation.context,
        next_function_to_call=blocking_operation.next_function_to_call,
    ), {"type": "instructor_completion_response", "response": response.model_dump()}



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from swarmstar.swarm.types import (
    ActionFolder,
    ActionSpace,
    BlockingOperation,
    NodeEmbryo,
    SpawnOperation,
    SwarmOperation,
)
from swarmstar.swarm.types.base_action import BaseAction


class NextActionPath(BaseModel):
    index: Optional[int] = Field(
        None, description="Index of the best action path to take"
    )
    failure_message: Optional[str] = Field(
        None,
        description="There's no good action path to take. Describe what type of action is needed in detail.",
    )


ROUTE_ACTION_INSTRUCTIONS = (
    "Decide what action path to take based on the goal and the available actions. "
    "If there is no good action path to take, describe what type of action is needed "
    "in detail in the failure message, and leave index empty."
)


class RouteAction(BaseAction):
    def main(self) -> BlockingOperation:
        action_space = ActionSpace(swarm=self.swarm)
        root: ActionFolder = action_space["swarmstar/actions"]
        root_children_descriptions = self.get_children_descriptions(action_space, root)
        messages = self.build_route_messages(
            self.node.message, root_children_descriptions
        )

        self.add_journal_entry(
            {
                "header": "Decideing what action to take given directive",
                "content": self.node.message
            }
        )

        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={"messages": messages, "instructor_model_name": "NextActionPath"},
            context={"parent_action_id": "swarmstar/actions"},
            next_function_to_call="route_goal",
        )

    def route_goal(
        self, completion: NextActionPath, parent_action_id: str
    ) -> SwarmOperation:
        """
        This function gets called over and over again 
        until we reach a leaf node, aka an action.
        """
        
        if completion.index is not None:
            action_space = ActionSpace(swarm=self.swarm)
            parent_action = action_space[parent_action_id]
            next_action_id = parent_action.children_ids[completion.index]
            current_action = action_space[next_action_id]

            if current_action.is_folder:
                children_descriptions = self.get_children_descriptions(
                    action_space, current_action
                )
                messages = self.build_route_messages(
                    self.node.message, children_descriptions
                )
                return BlockingOperation(
                    node_id=self.node.node_id,
                    blocking_type="instructor_completion",
                    args={
                        "messages": messages,
                        "instructor_model_name": "NextActionPath",
                    },
                    context={"parent_action_id": next_action_id},
                    next_function_to_call="route_goal",
                )
            else:
                self.add_journal_entry({
                    "header": "Successfully routed action",
                    "content": f"Routed goal: {self.node.message} to {next_action_id}"
                })
                return SpawnOperation(
                    node_id=self.node.node_id,
                    node_embryo=NodeEmbryo(
                        action_id=next_action_id, message=self.node.message
                    )
                )
        else:
            failure_message = completion.failure_message
            raise ValueError(
                f"The router agent failed to find a good action path to take. We need to implement something here to handle this. For example we could pass this to the action creator or talk to the user.\n\nThe agent's failure message: {failure_message}"
            )
            # TODO Handle failure message. Pass to action creator or user for review

    def build_route_messages(
        self, goal: str, children_descriptions: List[str]
    ) -> List[Dict[str, str]]:
        goal_and_action_path_options = (
            f"Decide what action path is best to take to accomplish this goal: {goal}\n\n"
            f"Options:\n"
        )
        for i, description in enumerate(children_descriptions):
            goal_and_action_path_options += f"{i}. {description}\n"

        messages = [
            {"role": "system", "content": ROUTE_ACTION_INSTRUCTIONS},
            {"role": "system", "content": goal_and_action_path_options},
        ]
        return messages

    def get_children_descriptions(
        self, action_space: ActionSpace, action_folder: ActionFolder
    ) -> List[str]:
        children_descriptions = []
        for child_id in action_folder.children_ids:
            child_metadata = action_space[child_id]
            children_descriptions.append(child_metadata.description)
        return children_descriptions



import os


def validate_and_adjust_swarm_space_path(path: str, root_path: str) -> dict:
    """
    Validates and adjusts the path based on the swarm space root path.

    1. If the path is directly under the root path, it's valid.
    2. If the path is relative to the root path, it's adjusted and returned.
    3. If the path is outside the allowed root space, an error is returned.

    Args:
    - path: The path to the file or folder.
    - root_path: The root path to validate against.

    Returns:
    - The adjusted path if valid.
    """
    # Normalize paths to ensure consistent comparisons
    root_path = os.path.normpath(root_path)
    path = os.path.normpath(path)

    # Ensure the root_path ends with a separator to accurately detect subdirectories
    if not root_path.endswith(os.path.sep):
        root_path += os.path.sep

    # Case 1: File path is directly under the root path
    if path.startswith(root_path):
        return path

    # Construct a full path assuming file_path is relative to root_path
    potential_path = os.path.join(root_path, path)
    potential_path = os.path.normpath(potential_path)

    # Case 2: After adjustment, if the potential path still starts with root_path, it's valid
    if potential_path.startswith(root_path):
        return potential_path

    # Case 3: File path is outside the allowed root space
    raise ValueError("File path is outside the allowed root space")



"""
    Parralel review termination is designated decompose directive actions.
    
    A node of this type decomposes its directive into immediate parralelizable actionable subdirectives.
    Thus it will have multiple children, and in addition it's goal might not be accomplished even when
    all it's children have terminated. We thus have a specific termination policy for node' of this type.
    
    1. If any children are alive, do nothing.
    2. If all children are terminated spawn the 'confirm_completion' node.
    3. If all children are terminated including a 'confirm_completion' node, spawn
    a 'consolidate_reports' node
    4. If all children are terminated including a 'consolidate_reports' node, append the
    'consolidate_reports' node's report to this node's reports and terminate.
"""

from typing import Union

from swarmstar.swarm.types import (
    NodeEmbryo,
    SpawnOperation,
    SwarmConfig,
    SwarmState,
    TerminationOperation,
)


def terminate(
    swarm: SwarmConfig, termination_operation: TerminationOperation
) -> Union[TerminationOperation, None]:
    node_id = termination_operation.node_id
    swarm_state = SwarmState(swarm=swarm)
    node = swarm_state[node_id]

    mission_completion = False
    reports_consolidated = False
    node_with_final_report = ""

    for child_id in node.children_ids:
        child = swarm_state[child_id]
        if child.alive:
            return None
        if child.action_id == "swarmstar/actions/reasoning/confirm_completion":
            mission_completion = True
        if child.action_id == "swarmstar/actions/reasoning/consolidate_reports":
            reports_consolidated = True
            node_with_final_report = swarm_state[child_id]

    if mission_completion == False:
        return SpawnOperation(
            operation_type="spawn",
            node_id=node_id,
            node_embryo=NodeEmbryo(
                node_id=node_id,
                action_id="swarmstar/actions/reasoning/confirm_completion",
                message="",
            ),
        )
    else:
        if reports_consolidated == False:
            return SpawnOperation(
                operation_type="spawn",
                node_id=node_id,
                node_embryos=[
                    NodeEmbryo(
                        node_id=node_id,
                        action_id="swarmstar/actions/reasoning/consolidate_reports",
                        message="",
                    )
                ],
            )
        else:
            node.report = node_with_final_report.report
            node.alive = False
            swarm_state.update_state(node)
            return TerminationOperation(
                operation_type="terminate",
                node_id=node_id,
            )



"""
This is a common interface for folder operations for the swarm space.

All folder operations will happen within the swarm space root path.
"""
from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, List

from swarmstar.utils.data.paths import validate_and_adjust_swarm_space_path

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig

platform_map = {
    "mac": "swarmstar.utils.data.folder_operations.local_storage",
    "azure": "swarmstar.utils.data.folder_operations.azure_blob_storage",
}


def delete_swarm_space_folder(swarm: SwarmConfig, folder_path: str) -> None:
    platform = swarm.platform
    folder_path = validate_and_adjust_swarm_space_path(
        folder_path, swarm.swarm_space_root_path
    )
    folder_operations_module = import_module(platform_map[platform])
    return folder_operations_module.delete_folder(swarm, folder_path)


def list_swarm_space_folder(swarm: SwarmConfig, folder_path: str) -> List[str]:
    platform = swarm.platform
    folder_path = validate_and_adjust_swarm_space_path(
        folder_path, swarm.swarm_space_root_path
    )
    folder_operations_module = import_module(platform_map[platform])
    return folder_operations_module.list_folder(swarm, folder_path)


def make_swarm_space_folder(swarm: SwarmConfig, folder_path: str) -> None:
    platform = swarm.platform
    folder_path = validate_and_adjust_swarm_space_path(
        folder_path, swarm.swarm_space_root_path
    )
    folder_operations_module = import_module(platform_map[platform])
    return folder_operations_module.make_folder(swarm, folder_path)


def move_swarm_space_folder(
    swarm: SwarmConfig, folder_path: str, new_folder_path: str
) -> None:
    platform = swarm.platform
    folder_path = validate_and_adjust_swarm_space_path(
        folder_path, swarm.swarm_space_root_path
    )
    new_folder_path = validate_and_adjust_swarm_space_path(
        new_folder_path, swarm.swarm_space_root_path
    )
    folder_operations_module = import_module(platform_map[platform])
    return folder_operations_module.move_folder(swarm, folder_path, new_folder_path)


def rename_swarm_space_folder(
    swarm: SwarmConfig, folder_path: str, new_folder_name: str
) -> None:
    platform = swarm.platform
    folder_path = validate_and_adjust_swarm_space_path(
        folder_path, swarm.swarm_space_root_path
    )
    new_folder_name = validate_and_adjust_swarm_space_path(
        new_folder_name, swarm.swarm_space_root_path
    )
    folder_operations_module = import_module(platform_map[platform])
    return folder_operations_module.rename_folder(swarm, folder_path, new_folder_name)



import importlib

from swarmstar.utils.retrieval.ast_chunking import find_class_start_end_lines
from swarmstar.utils.retrieval.get_lines import get_lines


def get_class_as_string(file_path: str, class_name: str) -> str:
    """
    Extracts the code for a class from a file.

    Parameters:
    - file_path: str - The path to the file from which to extract the code.
    - class_name: str - The name of the class to extract.

    Returns:
    - str - The extracted code as a string.
    """

    # Find the start and end lines of the class
    start_line, end_line = find_class_start_end_lines(file_path, class_name)

    # Extract the code for the class
    class_code = get_lines(file_path, (start_line, end_line))

    return class_code


def get_class_from_module(module_path: str, class_name: str):
    """
    Dynamically imports a class from a given module path and class name.

    :param module_path: A string representing the module path, relative to the package root, without the '.py' extension.
                        Example: 'my_package.my_module'
    :param class_name: The name of the class to import as a string.
    :return: The class type specified by class_name if found in the module; raises AttributeError otherwise.
    """
    try:
        # Dynamically import the module specified by module_path
        module = importlib.import_module(module_path)
        # Get the class specified by class_name from the imported module
        clazz = getattr(module, class_name)
        return clazz
    except ImportError as e:
        raise ImportError(f"Could not import module {module_path}: {e}") from None
    except AttributeError as e:
        raise AttributeError(
            f"Class {class_name} not found in module {module_path}: {e}"
        ) from None


# Example usage:
# Assuming there's a class MyClass in the module_path 'my_package.my_module'
# MyClass = get_class_from_module('my_package.my_module', 'MyClass')



from typing import List, Union

from swarmstar.swarm.types import SwarmConfig, SwarmOperation, SpawnOperation, NodeEmbryo
from swarmstar.utils.swarm_utils.blocking_operations.main import blocking
from swarmstar.utils.swarm_utils.failure_operations.main import failure
from swarmstar.utils.swarm_utils.spawn_operations.main import spawn
from swarmstar.utils.swarm_utils.termination_operations.main import terminate

class Swarmstar:
    def __init__(self, swarm: SwarmConfig):
        self.swarm = swarm
        
    def spawn_root(self, goal: str) -> [List[SpawnOperation], str]:
        """
        Create the first spawn operation for the swarm.
        """
        root_spawn_operation = SpawnOperation(
            node_embryo=NodeEmbryo(
                action_id='swarmstar/actions/reasoning/decompose_directive',
                message=goal
            )
        )
        return root_spawn_operation
    
    def execute(
        self, swarm_operation: SwarmOperation
    ) -> Union[List[SwarmOperation], None]:
        """
        This function is the main entry point for the swarmstar library. It takes in a swarm configuration and a swarm operation
        and returns a list of swarm operations that should be executed next.
        """
        operation_mapping = {
            "spawn": spawn,
            "blocking": blocking,
            "terminate": terminate,
            "node_failure": failure,
        }

        if swarm_operation.operation_type in operation_mapping:
            output = operation_mapping[swarm_operation.operation_type](
                self.swarm, swarm_operation
            )
        else:
            raise ValueError(
                f"Unknown swarm operation type: {swarm_operation.operation_type}"
            )

        if not isinstance(output, list) and output is not None:
            output = [output]
        return output



def get_lines(file_path, line_range):
    """
    Extracts code from a file given a range of line numbers.

    Parameters:
    - file_path: str - The path to the file from which to extract the code.
    - line_range: tuple(int, int) - A tuple containing the start and end line numbers.

    Returns:
    - str: The extracted code as a string.
    """
    start_line, end_line = line_range
    extracted_code = ""

    with open(file_path, "r") as file:
        for current_line_number, line in enumerate(file, start=1):
            if start_line <= current_line_number <= end_line:
                extracted_code += line
            elif current_line_number > end_line:
                break

    return extracted_code



"""
The memory metadata is a place for the swarm to store it's work,
as well as a place for it to perform various forms of RAG. 

First and foremost, the memory metadata organizes the memory so 
routers can search over the space effectively.

Secondly, data can take on many forms. Internal package documentation,
cloud blob storage, local file systems, vector databases and more.
All of these require different types of interaction. The memory 
metadata labels the memory so the swarm knows how to interact with it.
"""

from typing import Dict, List

from pydantic import BaseModel
from typing_extensions import Literal

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.utils.data.internal_operations import get_internal_memory_metadata
from swarmstar.utils.data.kv_operations.main import get_kv


class MemoryMetadata(BaseModel):
    type: Literal["internal_folder", "local_folder", "azure_blob"]
    name: str
    description: str
    parent: str
    children: List[str]
    metadata: Dict[str, str]


class MemorySpace(BaseModel):
    swarm: SwarmConfig

    def __getitem__(self, memory_id: str) -> MemoryMetadata:
        try:
            internal_memory_metadata = get_internal_memory_metadata(
                self.swarm, memory_id
            )
            return internal_memory_metadata
        except Exception:
            external_memory_metadata = get_kv(self.swarm, "memory_space", memory_id)
            if external_memory_metadata is not None:
                return external_memory_metadata
            else:
                raise ValueError(f"This memory id: `{memory_id}` does not exist.")



"""
This module contains the base class for all actions.
This base class: 

    - Provides some common methods that actions use.
    - Provides a metaclass to apply an error handling decorator to all methods of the action.
    
When a new action is created, it should subclass BaseAction and implement the main method.

All action functions will automatically be wrapped with the error handling decorator, which 
will catch any exceptions and return a FailureOperation with a report of the error.
"""
import traceback
from abc import ABCMeta, abstractmethod
from functools import wraps
from typing import Any, Dict

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.swarm.types.swarm_nodes import SwarmNode
from swarmstar.swarm.types.swarm_operations import SwarmOperation
from swarmstar.swarm.types.swarm_state import SwarmState


def error_handling_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            self = args[
                0
            ]  # Assuming the first argument is always 'self' for instance methods
            tb_str = traceback.format_exc()
            params_str = f"node_id: {self.node.node_id}\nParams: {kwargs}"

            error_message = (
                f"Error in {func.__name__}:\n{str(e)}\n\n{tb_str}\n\n{params_str}"
            )
            raise Exception(error_message)

            # return FailureOperation(
            #     node_id=self.node.node_id,
            #     report=report,
            # )

    return wrapper


class ErrorHandlingMeta(ABCMeta):
    def __new__(mcs, name, bases, dct):
        new_cls = super().__new__(mcs, name, bases, dct)
        for attr_name, attr_value in dct.items():
            if callable(attr_value) and not attr_name.startswith("__"):
                error_wrapped = error_handling_decorator(attr_value)
                setattr(new_cls, attr_name, error_wrapped)
        return new_cls

class BaseAction(metaclass=ErrorHandlingMeta):
    """
    All actions should subclass this class.
    """

    def __init__(self, swarm: SwarmConfig, node: SwarmNode):
        self.swarm = swarm
        self.node = node

    @abstractmethod
    def main(self, **kwargs) -> SwarmOperation:
        pass

    def add_journal_entry(self, journal_entry: Dict[str, Any]):
        self.node.journal.append(journal_entry)
        swarm_state = SwarmState(swarm=self.swarm)
        swarm_state.update_state(self.node)

    def add_developer_log(self, developer_log: Dict[str, Any]):
        self.node.developer_logs.append(developer_log)
        swarm_state = SwarmState(swarm=self.swarm)
        swarm_state.update_state(self.node)

    def update_termination_policy(self, termination_policy: str):
        self.node.termination_policy = termination_policy
        swarm_state = SwarmState(swarm=self.swarm)
        swarm_state.update_state(self.node)



import uuid
from datetime import datetime


def generate_uuid(identifier):
    if not isinstance(identifier, str):
        raise TypeError("Identifier must be a string")

    timestamp = datetime.now().strftime("%Y%m%d%H%M%S%f")
    unique_id = uuid.uuid4()
    return f"{identifier}_{timestamp}_{unique_id}"



from __future__ import annotations

import subprocess
from importlib import resources
from typing import TYPE_CHECKING

import pymongo
from pymongo import MongoClient

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def check_and_create_database(mongodb_uri: str, db_name: str) -> None:
    """
    Check if a MongoDB database exists, and if not, create it.

    :param mongodb_uri: MongoDB URI for connecting to the instance.
    :param db_name: Name of the database to check and potentially create.
    :raises ValueError: If the database already exists.
    """
    client = MongoClient(mongodb_uri)
    if db_name in client.list_database_names():
        raise ValueError(f"Database {db_name} already exists.")
    else:
        db = client[db_name]
        collection = db["stage"]
        collection.insert_one({"_id": "dummy_id"})
        collection.delete_one({"_id": "dummy_id"})
        print(f"Database {db_name} created successfully.")


def restore_database(package_name: str, resource_name: str, uri: str, db_name: str):
    """
    Restore a MongoDB database dump to a specified database from package resources.

    :param package_name: Name of the package containing the dump files.
    :param resource_name: The resource name/path within the package.
    :param uri: MongoDB URI for connecting to the target instance.
    :param db_name: Name of the database to restore the dump into.
    """
    # Use importlib.resources to resolve the path to the resource
    with resources.path(package_name, resource_name) as dump_path:
        # Construct the mongorestore command
        restore_command = ["mongorestore", "--uri", uri, "--db", db_name, str(dump_path)]

        try:
            # Execute the mongorestore command
            subprocess.run(restore_command, check=True)
            print("Database restored successfully.")
        except subprocess.CalledProcessError as e:
            raise ValueError(f"Failed to restore database: {e}")


def create_client(uri: str) -> MongoClient:
    try:
        client = MongoClient(uri)
        return client
    except Exception as e:
        raise ValueError(f"Failed to create MongoDB client: {str(e)}")


def create_collection(swarm: SwarmConfig, category: str) -> None:
    uri = swarm.mongodb_uri
    db_name = swarm.mongodb_db_name
    client = create_client(uri)
    db = client[db_name]
    collection = db[category]
    collection.create_index([("key", pymongo.ASCENDING)], unique=True)


def add_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    try:
        db_name = swarm.mongodb_db_name
        collection_name = category
        uri = swarm.mongodb_uri
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        document = {"key": key, **value}
        collection.insert_one(document)
    except pymongo.errors.DuplicateKeyError as e:
        update_kv(swarm, category, key, value)
    except Exception as e:
        raise ValueError(f"Failed to add to MongoDB collection: {str(e)}")
    
def add_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    """
    Add a key-value pair to the collection with an initial version number.
    """
    try:
        db_name = swarm.mongodb_db_name
        collection_name = category
        uri = swarm.mongodb_uri
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        # Initialize the document with a version number
        document = {"key": key, "version": 1, **value}
        collection.insert_one(document)
    except pymongo.errors.DuplicateKeyError:
        update_kv(swarm, category, key, value)
    except Exception as e:
        raise ValueError(f'Failed to add to MongoDB collection: {str(e)}')


def get_kv(swarm: SwarmConfig, category: str, key: str) -> dict:
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        result = collection.find_one({"key": key})
        if result is None:
            raise ValueError(f"Key {key} not found in MongoDB collection.")
        result.pop("_id")
        result.pop("key")
        result.pop("version")
        return result
    except Exception as e:
        raise ValueError(f"Failed to get from MongoDB collection: {str(e)}")


def delete_kv(swarm: SwarmConfig, category: str, key: str) -> None:
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        result = collection.delete_one({"key": key})
        if result.deleted_count == 0:
            raise ValueError(f"Key {key} not found in MongoDB collection.")
    except Exception as e:
        raise ValueError(f"Failed to delete from MongoDB collection: {str(e)}")

def update_kv(swarm: SwarmConfig, category: str, key: str, updated_values: dict) -> None:
    """
    Update specified fields of a document and increment its version, ensuring optimistic concurrency control.
    """
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        collection = db[collection_name]
        
        retries = 3
        for attempt in range(retries):
            # Fetch the current document
            current_document = collection.find_one({"key": key})
            if current_document is None:
                raise ValueError(f'Key {key} not found in MongoDB collection {collection_name}.')

            # Prepare the update
            new_version = current_document.get("version", 0) + 1
            update_fields = {"version": new_version}
            for field, value in updated_values.items():
                if field not in current_document:
                    raise KeyError(f"Field '{field}' does not exist in the document.")
                update_fields[field] = value

            # Attempt to update the document if the version hasn't changed
            result = collection.update_one(
                {"key": key, "version": current_document["version"]},
                {"$set": update_fields}
            )

            if result.matched_count:
                break  # Update was successful
            elif attempt == retries - 1:
                raise Exception("Failed to update document due to concurrent modification.")
            # If the document was updated elsewhere, retry the operation
    except Exception as e:
        raise ValueError(f"Failed to update MongoDB collection: {str(e)}")

def append_to_list_with_versioning(swarm: SwarmConfig, category: str, key: str, list_field: str, value_to_append) -> None:
    """
    Appends a value to a list within a document, using optimistic concurrency control to handle concurrent updates safely.

    :param collection_name: The name of the collection.
    :param key: The key identifying the document.
    :param list_field: The field within the document that contains the list to append to.
    :param value_to_append: The value to append to the list.
    """
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        collection = db[collection_name]

        retries = 3
        for attempt in range(retries):
            # Fetch the current document
            current_document = collection.find_one({"key": key})
            if current_document is None:
                raise ValueError(f'Key {key} not found in MongoDB collection {collection_name}.')
            if list_field not in current_document:
                raise KeyError(f"Field '{list_field}' does not exist in the document. Ensure the field is initialized as a list.")

            # Prepare the update with an incremented version number
            new_version = current_document.get("version", 0) + 1
            updated_values = {"$push": {list_field: value_to_append}, "$set": {"version": new_version}}

            # Attempt to update the document if the version hasn't changed
            result = collection.update_one(
                {"key": key, "version": current_document["version"]},
                updated_values
            )

            if result.matched_count:
                break  # Update was successful
            elif attempt == retries - 1:
                raise Exception("Failed to update document due to concurrent modification.")
            # If the document was updated elsewhere, retry the operation
    except Exception as e:
        raise ValueError(f"Failed to append to list in MongoDB collection: {str(e)}")

def add_to_dict_with_versioning(swarm: SwarmConfig, category: str, key: str, dict_field: str, dict_key: str, dict_value) -> None:
    """
    Adds a key-value pair to a dictionary within a document, using optimistic concurrency control to handle concurrent updates safely.

    :param collection_name: The name of the collection.
    :param key: The key identifying the document.
    :param dict_field: The field within the document that contains the dictionary to add to.
    :param dict_key: The key to add to the dictionary.
    :param dict_value: The value to associate with the dict_key in the dictionary.
    """
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        collection = db[collection_name]

        retries = 3
        for attempt in range(retries):
            # Fetch the current document
            current_document = collection.find_one({"key": key})
            if current_document is None:
                raise ValueError(f'Key {key} not found in MongoDB collection {collection_name}.')
            if dict_field not in current_document:
                raise KeyError(f"Field '{dict_field}' does not exist in the document. Ensure the field is initialized as a dictionary.")

            # Prepare the update with an incremented version number
            new_version = current_document.get("version", 0) + 1
            updated_values = {
                "$set": {
                    f"{dict_field}.{dict_key}": dict_value,  # This adds or updates the key in the dictionary
                    "version": new_version
                }
            }

            # Attempt to update the document if the version hasn't changed
            result = collection.update_one(
                {"key": key, "version": current_document["version"]},
                updated_values
            )

            if result.matched_count:
                break  # Update was successful
            elif attempt == retries - 1:
                raise Exception("Failed to update document due to concurrent modification.")
            # If the document was updated elsewhere, retry the operation
    except Exception as e:
        raise ValueError(f"Failed to add to dictionary in MongoDB collection: {str(e)}")


import ast


def find_class_start_end_lines(file_path, class_name):
    """Find the start and end lines of a class in the given file."""

    # Read the source code from the file
    with open(file_path, "r") as file:
        source_code = file.read()

    # Parse the source code into an AST
    tree = ast.parse(source_code)

    # Define a class visitor to find the specific class
    class ClassVisitor(ast.NodeVisitor):
        def __init__(self):
            self.start_line = None
            self.end_line = None

        def visit_ClassDef(self, node):
            """Visit a class definition."""
            if node.name == class_name:
                # Found the class, get its start line
                self.start_line = node.lineno
                # Python 3.8 and later: directly use end_lineno if available
                self.end_line = getattr(node, "end_lineno", None)

                # For Python versions < 3.8, find the end line by looking at the last body element
                if self.end_line is None and node.body:
                    # The last element of the class body might not be the actual last line (e.g., comments),
                    # but it's a reasonable approximation without parsing comments.
                    self.end_line = node.body[-1].lineno

                # Stop visiting more nodes
                raise StopIteration

    # Create the visitor and visit the AST nodes
    visitor = ClassVisitor()
    try:
        visitor.visit(tree)
    except StopIteration:
        pass  # Expected behavior to stop after finding the class

    # Return the start and end lines
    return visitor.start_line, visitor.end_line


# def parse_file(file_path):
#     with open(file_path, 'r') as file:
#         return ast.parse(file.read(), filename=file_path)

# class ParseCodeAction:
#     def execute(self, file_path):
#         tree = parse_file(file_path)
#         # Further processing to identify classes and functions.
#         return tree

# class IdentifyDependenciesAction:
#     def execute(self, tree):
#         dependencies = {'imports': [], 'internal_references': []}
#         for node in ast.walk(tree):
#             if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
#                 for alias in node.names:
#                     dependencies['imports'].append(alias.name)
#             # Add logic for identifying internal references.
#         return dependencies


# class GatherCodeAction:
#     def execute(self, root_path, dependencies):
#         all_code = {}
#         for dep in dependencies['imports']:
#             # Convert import to file path. This may require custom logic.
#             file_path = convert_import_to_path(root_path, dep)
#             if file_path:
#                 tree = parse_file(file_path)
#                 all_code[file_path] = tree
#                 # Recursively gather dependencies.
#                 sub_deps = IdentifyDependenciesAction().execute(tree)
#                 all_code.update(self.execute(root_path, sub_deps))
#         # Handle internal_references similarly.
#         return all_code

# def analyze_codebase(entry_file_path):
#     parse_action = ParseCodeAction()
#     identify_deps_action = IdentifyDependenciesAction()
#     gather_code_action = GatherCodeAction()

#     tree = parse_action.execute(entry_file_path)
#     dependencies = identify_deps_action.execute(tree)
#     all_related_code = gather_code_action.execute(os.path.dirname(entry_file_path), dependencies)

#     # all_related_code contains the ASTs for all dependencies.
#     # Further processing can be done to format, display, or analyze this code.


# def parse_file(file_path):
#     with open(file_path, 'r') as file:
#         return ast.parse(file.read(), filename=file_path)

# class ParseCodeAction:
#     def execute(self, file_path):
#         tree = parse_file(file_path)
#         # Further processing to identify classes and functions.
#         return tree

# class IdentifyDependenciesAction:
#     def execute(self, tree):
#         dependencies = {'imports': [], 'internal_references': []}
#         for node in ast.walk(tree):
#             if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
#                 for alias in node.names:
#                     dependencies['imports'].append(alias.name)
#             # Add logic for identifying internal references.
#         return dependencies

# def convert_import_to_path(root_path, import_name):
#     # Custom logic to convert an import to a file path.
#     # This may involve searching directories, resolving package names, etc.
#     pass # TODO: Implement this logic

# class GatherCodeAction:
#     def execute(self, root_path, dependencies):
#         all_code = {}
#         for dep in dependencies['imports']:
#             # Convert import to file path. This may require custom logic.
#             file_path = convert_import_to_path(root_path, dep)
#             if file_path:
#                 tree = parse_file(file_path)
#                 all_code[file_path] = tree
#                 # Recursively gather dependencies.
#                 sub_deps = IdentifyDependenciesAction().execute(tree)
#                 all_code.update(self.execute(root_path, sub_deps))
#         # Handle internal_references similarly.
#         return all_code

# def analyze_codebase(entry_file_path):
#     parse_action = ParseCodeAction()
#     identify_deps_action = IdentifyDependenciesAction()
#     gather_code_action = GatherCodeAction()

#     tree = parse_action.execute(entry_file_path)
#     dependencies = identify_deps_action.execute(tree)
#     all_related_code = gather_code_action.execute(os.path.dirname(entry_file_path), dependencies)

# all_related_code contains the ASTs for all dependencies.
# Further processing can be done to format, display, or analyze this code.


# Challenges and Optimization:
# Path Resolution: Converting imports to file paths is specific to your project's structure and may require custom logic.
# Performance: Recursively parsing and gathering code can be resource-intensive. Consider optimizations like caching parsed files.
# Circular Dependencies: Implement checks to avoid infinite recursion in case of circular dependencies.



from importlib import import_module
from typing import Union

from swarmstar.swarm.types import (
    SwarmConfig,
    SwarmHistory,
    SwarmState,
    TerminationOperation,
)


def terminate(
    swarm: SwarmConfig, termination_operation: TerminationOperation
) -> Union[TerminationOperation, None]:
    termination_policy_map = {
        "simple": "swarmstar.utils.swarm_utils.termination_operations.simple",
        "parallel_review": "swarmstar.utils.swarm_utils.termination_operations.parallel_review",
        "clone_with_questions_answered": "swarmstar.utils.swarm_utils.termination_operations.clone_with_questions_answered",
    }

    swarm_state = SwarmState(swarm=swarm)
    node_id = termination_operation.node_id
    node = swarm_state[node_id]
    termination_policy = node.termination_policy

    if termination_policy not in termination_policy_map:
        raise ValueError(
            f"Termination policy: `{termination_policy}` is not supported."
        )
    else:
        pass

    termination_policy_module = import_module(
        termination_policy_map[termination_policy]
    )

    output = termination_policy_module.terminate(swarm, termination_operation)
    swarm_history = SwarmHistory(swarm=swarm)
    swarm_history.add_event(termination_operation)

    return output



"""
    Simple termination simply terminates the given node and returns a TerminationOperation for the parent node.
"""
from typing import Union

from swarmstar.swarm.types import SwarmConfig, SwarmState, TerminationOperation
from swarmstar.utils.ai.openai_instructor import completion

def terminate(
    swarm: SwarmConfig, termination_operation: TerminationOperation
) -> Union[TerminationOperation, None]:
    node_id = termination_operation.node_id
    swarm_state = SwarmState(swarm=swarm)
    node = swarm_state[node_id]
    node.alive = False
    swarm_state.update_state(node)

    try:
        parent_node = swarm_state[node.parent_id]
    except:
        return None

    return TerminationOperation(
        node_id=parent_node.node_id, 
    )



'''
When the swarm object is created a single sqlite3 database is created for the swarm space.

This database is used to store all kv store data with the key being the category and the key.
'''
from __future__ import annotations
from typing import TYPE_CHECKING
import sqlite3
import json

from swarmstar.swarm.types import SwarmConfig

def create_or_open_kv_db(db_path: str) -> None:
    try:
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute('CREATE TABLE IF NOT EXISTS kv_store (key TEXT PRIMARY KEY, value TEXT)')
        conn.commit()
    except Exception as e:
        raise ValueError(f'Failed to create or open kv store db: {str(e)}')

def upload_swarm_space_kv_pair(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    try:
        conn = sqlite3.connect(swarm.configs.sqlite3_db_path)
        cursor = conn.cursor()
        key = f'{category}_{key}'
        if type(value) != str: value = json.dumps(value)        
        cursor.execute('INSERT OR REPLACE INTO kv_store (key, value) VALUES (?, ?)', (key, value))
        conn.commit()
    except Exception as e:
        raise ValueError(f'Failed to upload kv pair: {str(e)}')

def retrieve_swarm_space_kv_value(swarm: SwarmConfig, category: str, key: str) -> dict:
    try:
        conn = sqlite3.connect(swarm.configs.sqlite3_db_path)
        cursor = conn.cursor()
        key = f'{category}_{key}'
        cursor.execute('SELECT value FROM kv_store WHERE key = ?', (key,))
        result = cursor.fetchone()
        if result:
            
            return json.loads(result[0])
        raise ValueError(f'No value found for key: {key}')
    except Exception as e:
        raise ValueError(f'Failed to retrieve kv value: {str(e)} at {swarm.configs.sqlite3_db_path}')

def delete_swarm_space_kv_pair(swarm: SwarmConfig, category: str, key: str) -> None:
    try:
        conn = sqlite3.connect(swarm.configs.sqlite3_db_path)
        cursor = conn.cursor()
        key = f'{category}_{key}'
        cursor.execute('DELETE FROM kv_store WHERE key = ?', (key,))
        conn.commit()
    except Exception as e:
        raise ValueError(f'Failed to delete kv pair: {str(e)}')



from typing import List, Union

from swarmstar.swarm.types import FailureOperation, SwarmConfig, SwarmOperation


def failure(
    swarm: SwarmConfig, node_failure_operation: FailureOperation
) -> Union[SwarmOperation, List[SwarmOperation]]:
    raise (node_failure_operation.report)
    # TODO Handle failure operation



"""
This is a common interface for file operations for the swarm space.

All file operations will happen within the swarm space root path.
"""
from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING

from swarmstar.utils.data.paths import validate_and_adjust_swarm_space_path

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig

platform_map = {
    "mac": "swarmstar.utils.data.file_operations.local_storage",
    "azure": "swarmstar.utils.data.file_operations.azure_blob_storage",
}


def delete_swarm_space_file(swarm: SwarmConfig, file_path: str) -> None:
    platform = swarm.platform
    file_path = validate_and_adjust_swarm_space_path(
        file_path, swarm.swarm_space_root_path
    )
    file_operations_module = import_module(platform_map[platform])
    return file_operations_module.delete_file(swarm, file_path)


def move_swarm_space_file(
    swarm: SwarmConfig, file_path: str, new_file_path: str
) -> None:
    platform = swarm.platform
    file_path = validate_and_adjust_swarm_space_path(
        file_path, swarm.swarm_space_root_path
    )
    new_file_path = validate_and_adjust_swarm_space_path(
        new_file_path, swarm.swarm_space_root_path
    )
    file_operations_module = import_module(platform_map[platform])
    return file_operations_module.move_file(swarm, file_path, new_file_path)


def rename_swarm_space_file(
    swarm: SwarmConfig, file_path: str, new_file_name: str
) -> None:
    platform = swarm.platform
    file_path = validate_and_adjust_swarm_space_path(
        file_path, swarm.swarm_space_root_path
    )
    new_file_name = validate_and_adjust_swarm_space_path(
        new_file_name, swarm.swarm_space_root_path
    )
    file_operations_module = import_module(platform_map[platform])
    return file_operations_module.rename_file(swarm, file_path, new_file_name)


def upload_swarm_space_file(
    swarm: SwarmConfig, file_path: str, file_bytes: bytes
) -> None:
    platform = swarm.platform
    file_path = validate_and_adjust_swarm_space_path(
        file_path, swarm.swarm_space_root_path
    )
    file_operations_module = import_module(platform_map[platform])
    return file_operations_module.upload_file(swarm, file_path, file_bytes)


def retrieve_swarm_space_file(swarm: SwarmConfig, file_path: str) -> bytes:
    platform = swarm.platform
    file_path = validate_and_adjust_swarm_space_path(
        file_path, swarm.swarm_space_root_path
    )
    file_operations_module = import_module(platform_map[platform])
    return file_operations_module.retrieve_file(swarm, file_path)



from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def delete_folder(swarm: SwarmConfig, folder_path: str) -> dict:
    try:
        shutil.rmtree(folder_path)
    except Exception as e:
        raise ValueError(f"Failed to delete folder: {str(e)}")


def list_folder(swarm: SwarmConfig, folder_path: str) -> List[str]:
    try:
        entries = os.listdir(folder_path)
        full_paths = [os.path.join(folder_path, entry) for entry in entries]
        return full_paths
    except Exception as e:
        raise ValueError(f"Failed to list folder: {str(e)}")


def make_folder(swarm: SwarmConfig, folder_path: str) -> dict:
    try:
        os.makedirs(folder_path, exist_ok=True)
    except Exception as e:
        raise ValueError(f"Failed to make folder: {str(e)}")


def move_folder(swarm: SwarmConfig, old_folder_path: str, new_folder_path: str) -> dict:
    try:
        shutil.move(old_folder_path, new_folder_path)
    except Exception as e:
        raise ValueError(f"Failed to move folder: {str(e)}")


def rename_folder(swarm: SwarmConfig, folder_path: str, new_folder_name: str) -> dict:
    try:
        os.rename(folder_path, new_folder_name)
    except Exception as e:
        raise ValueError(f"Failed to rename folder: {str(e)}")



"""
This module contains functions to load resources from within the package.
"""
import json
import sqlite3
from importlib import resources
from typing import Any, BinaryIO

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.utils.data.kv_operations.main import get_kv

def get_internal_action_metadata(action_id: str) -> dict:
    return get_internal_sqlite_kv("action_space", action_id)


def get_internal_memory_metadata(memory_id: str) -> dict:
    return get_internal_sqlite_kv("memory_space", memory_id)


def get_internal_util_metadata(util_id: str) -> dict:
    return get_internal_sqlite_kv("util_space", util_id)


def get_internal_sqlite_kv(category: str, key: str) -> dict:
    db_path = f'swarmstar/internal_metadata/{category}.sqlite3'
    try:
        with sqlite3.connect(db_path) as conn:
            cursor = conn.cursor()
            key = f'{category}_{key}'
            cursor.execute('SELECT value FROM kv_store WHERE key = ?', (key,))
            result = cursor.fetchone()
            if result:
                return json.loads(result[0])
            else:
                raise ValueError(f'No value found for key: {key}')
    except Exception as e:
        raise ValueError(f'Failed to retrieve kv value: {str(e)} at {db_path}')

def get_json_data(package: str, resource_name: str) -> Any:
    with resources.open_text(package, resource_name) as file:
        return json.load(file)


def get_binary_data(package: str, resource_name: str) -> bytes:
    with resources.open_binary(package, resource_name) as file:
        return file.read()


def get_binary_file(package: str, resource_name: str) -> BinaryIO:
    return resources.open_binary(package, resource_name)



"""
The actions that create blocking operations using instructor need to pass
the pydantic model. However, the pydantic model can't be serialized and 
deserialized. (There are serialization methods but no common deserialization methods)

So instead we'll store all the pydantic models in this file and import them, 
passing the name of the model as a string in the blocking operation.
"""

from typing import List, Optional

from pydantic import BaseModel, Field


class DecomposeDirectiveModel(BaseModel):
    scrap_paper: Optional[str] = Field(
        None,
        description="Scrap paper for notes, planning etc. Use this space to think step by step. (optional)",
    )
    questions: Optional[List[str]] = Field(
        None, description="Questions you need answered before decomposition."
    )
    subdirectives: Optional[List[str]] = Field(
        None,
        description="List of subdirectives to be executed in parallel, if you have no questions.",
    )


class NextActionPath(BaseModel):
    index: Optional[int] = Field(
        None, description="Index of the best action path to take"
    )
    failure_message: Optional[str] = Field(
        None,
        description="There's no good action path to take. Describe what type of action is needed in detail.",
    )


class QuestionAskerConversationState(BaseModel):
    questions: List[str] = Field(..., description="List of questions we need answered")
    persisted_context: str = Field(
        ...,
        description="A concise and compact representation of the necessary information to persist through the conversation.",
    )
    report: str = Field(
        ...,
        description="Concise and comprehensive report of answers to our questions and supporting context.",
    )


class AgentMessage(BaseModel):
    content: str = Field(..., description="The message to send to the user.")


class FinalReport(BaseModel):
    report: str = Field(
        ...,
        description="Concise and comprehensive report of answers to our questions and supporting context.",
    )


class ConfirmCompletionModel(BaseModel):
    is_complete: bool = Field(
        ..., description="Whether the directive is complete or not."
    )
    message: Optional[str] = Field(
        None, description="If the directive is not complete, this message will be sent back to the decompose directive node."
    )


"""
Every node in the swarm corresponds to an action.

The action space metadata organizes the actions and is used to search
for actions and execute them.

Every action has the same IO: 
    message: str, swarm: Swarm and node_id: str

The actual actions are all prewritten code. The action space metadata's
purpose is:

    (1) To allow the swarm to organize and search for actions
    (2) To allow the swarm to execute actions

Because the swarm can dynamically create actions we need (2). For example,
we might need to dynamically spin up a docker container and execute a function.
The execution_metadata provides a place to store the information to be passed
to the executor who handles this action_type.
"""
from typing import List, Optional

from pydantic import BaseModel, Field
from typing_extensions import Literal

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.utils.data.internal_operations import get_internal_action_metadata
from swarmstar.utils.data.kv_operations.main import get_kv


class ActionNode(BaseModel):
    is_folder: bool
    type: Literal[
        "azure_blob_storage_folder",
        "internal_folder",
        "internal_action",
        "azure_blob_action",
    ]
    name: str
    description: str
    children_ids: Optional[List[str]] = None
    parent: Optional[str] = None


class ActionFolder(ActionNode):
    is_folder: Literal[True] = Field(default=True)
    type: Literal["internal_folder", "azure_blob_storage_folder"]
    name: str
    description: str
    children_ids: List[str]
    parent: Optional[str] = None


class Action(ActionNode):
    is_folder: Literal[False] = Field(default=False)
    type: Literal["internal_action", "azure_blob_action"]
    name: str
    description: str
    parent: str
    children_ids: Optional[List[str]] = Field(default=None)
    termination_policy: Literal["simple", "parallel_review", "clone_with_questions_answered"]


class InternalAction(Action):
    type: Literal["internal_action"] = Field(default="internal_action")
    name: str
    description: str
    children_ids: Optional[List[str]] = Field(default=None)
    parent: str
    termination_policy: Literal["simple", "parallel_review", "clone_with_questions_answered"]
    internal_action_path: str


class InternalFolder(ActionFolder):
    type: Literal["internal_folder"] = Field(default="internal_folder")
    name: str
    description: str
    children_ids: List[str]
    parent: str = None
    internal_folder_path: str


class ActionSpace(BaseModel):
    """
    The action space metadata is stored in the swarm's kv store as:

        action_id: ActionMetadata
    """

    swarm: SwarmConfig

    def __getitem__(self, action_id: str) -> ActionNode:
        try:
            action_metadata = get_internal_action_metadata(self.swarm, action_id)
            if action_metadata is None:
                raise ValueError(
                    f"This action id: `{action_id}` does not exist in internal action space."
                )
        except Exception as e1:
            try:
                action_metadata = get_kv(self.swarm, "action_space", action_id)
                if action_metadata is None:
                    raise ValueError(
                        f"This action id: `{action_id}` does not exist in external action space."
                    ) from e1
            except Exception as e2:
                raise ValueError(
                    f"This action id: `{action_id}` does not exist in both internal and external action spaces."
                ) from e2

        type_mapping = {
            "internal_action": InternalAction,
            "internal_folder": InternalFolder,
        }
        action_type = action_metadata["type"]
        if action_type in type_mapping:
            return type_mapping[action_type](**action_metadata)
        return ActionNode(**action_metadata)

    def get_root(self) -> ActionNode:
        return self["swarmstar/actions"]



"""
A single container is used for all operations for all users.

CosmosDB automatically splits the container into physical partitions and does all the scaling.
We just need to choose 3 logical partitions and have an id field for retrieval.

The swarm object will hold 2 of the partiton keys and the keys to get the swarm container.

Partition keys: user_id/swarm_id/category
"category" can be "action_space_metadata", "util_space_metadata", "memory_space_metadata", "swarm_state" etc.

The "id" field will be the key for the document.
"id" might be like "action_id", "util_id", "memory_id", "node_id" etc.

In this manner we have good logical partitions and can provide a standard interface for all KV operations.

For now cosmosdb python sdk doesen't support hierarchical partitions. So we'll append all the
partition keys into a single string and use it as the partition key and id for now
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from azure.cosmos import CosmosClient

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def get_container(swarm: SwarmConfig, category: str):
    try:
        url = swarm.azure_cosmos_db_url
        cosmos_key = swarm.azure_cosmos_db_key
        container_name = swarm.azure_cosmos_db_container_name
        client = CosmosClient(url, credential=cosmos_key)
        database_name = client.get_database_client(swarm.azure_cosmos_db_database_name)
        container = database_name.get_container_client(container_name)
        return container
    except Exception as e:
        raise ValueError(f"Failed to connect to cosmosdb: {str(e)}") from e


def build_id(swarm: SwarmConfig, category: str, key: str) -> str:
    return f"{swarm.user_id}_{swarm.swarm_id}_{category}_{key}"


def add_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    container = get_container(swarm, category)

    id = build_id(swarm, category, key)
    value["id"] = id

    try:
        container.upsert_item(value)
    except Exception as e:
        raise ValueError(f"Failed to upload to cosmosdb: {str(e)}") from e


def get_kv(swarm: SwarmConfig, category: str, key: str):
    container = get_container(swarm, category)
    id = build_id(swarm, category, key)
    try:
        document = container.read_item(item=id, partition_key=id)
        return document
    except Exception as e:
        raise ValueError(f"Failed to retrieve from cosmosdb: {str(e)}") from e


def delete_kv(swarm: SwarmConfig, category: str, key: str):
    container = get_container(swarm, category)
    id = build_id(swarm, category, key)

    try:
        container.delete_item(item=id, partition_key=id)
    except Exception as e:
        raise ValueError(f"Failed to delete from cosmosdb: {str(e)}") from e


def update_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    container = get_container(swarm, category)
    build_id(swarm, category, key)

    try:
        container.upsert_item(value)
    except Exception as e:
        raise ValueError(f"Failed to update in cosmosdb: {str(e)}") from e


"""
The following works with a container that uses hierarchical partitions.
/ Partition keys: user_id/swarm_id/category
with id as the key for the document.
"""
# def add_kv(swarm: Swarm, category: str, key: str, value: dict):
#     url = swarm.azure_cosmos_db_url
#     cosmos_key = swarm.azure_cosmos_db_key
#     container_name = swarm.azure_cosmos_db_container_name
#     client = CosmosClient(url, credential=cosmos_key)
#     database_name = client.get_database_client(swarm.azure_cosmos_db_database_name)
#     container = database_name.get_container_client(container_name)

#     # Add partiton keys and id field. CosmosDB expects all values to be inside the dict.
#     value['id'] = key
#     value['category'] = category
#     value['user_id'] = swarm.user_id
#     value['swarm_id'] = swarm.swarm_id

#     try:
#         container.upsert_item(value)
#         return {'success': True, 'error_message': ''}
#     except Exception as e:
#         return {'success': False, 'error_message': str(e)}

# def get_kv(swarm: Swarm, category: str, key: str):
#     url = swarm.azure_cosmos_db_url
#     cosmos_key = swarm.azure_cosmos_db_key
#     container_name = swarm.azure_cosmos_db_container_name
#     client = CosmosClient(url, credential=cosmos_key)
#     database_name = client.get_database_client(swarm.azure_cosmos_db_database_name)
#     container = database_name.get_container_client(container_name)
#     partition_key = f'{swarm.user_id}/{swarm.swarm_id}/{category}'
#     partition_key = PartitionKey(partition_key)

#     try:
#         document = container.read_item(item=key, partition_key=key)
#         return {'success': True, 'error_message': '', 'data': document}
#     except Exception as e:
#         return {'success': False, 'error_message': str(e)}

# def delete_kv(swarm: Swarm, category: str, key: str):
#     url = swarm.azure_cosmos_db_url
#     cosmos_key = swarm.azure_cosmos_db_key
#     container_name = swarm.azure_cosmos_db_container_name
#     client = CosmosClient(url, credential=cosmos_key)
#     database_name = client.get_database_client(swarm.azure_cosmos_db_database_name)
#     container = database_name.get_container_client(container_name)
#     partition_key = f'{swarm.user_id}/{swarm.swarm_id}/{category}'
#     partition_key = PartitionKey(partition_key)

#     try:
#         container.delete_item(item=key, partition_key=key)
#         return {'success': True, 'error_message': ''}
#     except Exception as e:
#         return {'success': False, 'error_message': str(e)}



"""
Decompose a directive into actionable subdirectives.

The agent will ask questions if it needs more information before decomposing the directive.
"""
from typing import List, Optional

from pydantic import BaseModel, Field

from swarmstar.swarm.types import BlockingOperation, NodeEmbryo, SpawnOperation
from swarmstar.swarm.types.base_action import BaseAction


class DecomposeDirectiveModel(BaseModel):
    scrap_paper: Optional[str] = Field(
        None,
        description="Scrap paper for notes, planning etc. Use this space to think step by step. (optional)",
    )
    questions: Optional[List[str]] = Field(
        ..., description="Questions you need answered before decomposition."
    )
    subdirectives: Optional[List[str]] = Field(
        ...,
        description="List of subdirectives to be executed in parallel, if you have no questions.",
    )


DECOMPOSE_DIRECTIVE_INSTRUCTIONS = (
    "You are given a directive. You have 2 options:\n"
    "1. Ask questions to get more information or clarification of requirements and intentions.\n"
    "2. Decompose the directive into actionable subdirectives that will be executed independently and in parallel. "
    "After those are done, youll generate the next set of subdirectives. I stress that the subdirectives "
    "must be independent and parallel.\n\nChoose one of the options and proceed. Do not ask questions and decompose the directive at the same time."
)


class DecomposeDirective(BaseAction):
    def main(self) -> BlockingOperation:
        messages = [
            {"role": "system", "content": DECOMPOSE_DIRECTIVE_INSTRUCTIONS},
            {
                "role": "user",
                "content": f"Directive to decompose: \n`{self.node.message}`",
            },
        ]

        self.add_journal_entry(
            {
                "header": "Directive to decompose",
                "content": self.node.message,
            }
        )

        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={
                "messages": messages,
                "instructor_model_name": "DecomposeDirectiveModel",
            },
            context={},
            next_function_to_call="analyze_output",
        )

    def analyze_output(self, completion: DecomposeDirectiveModel) -> SpawnOperation:
        '''
        Depending on the completion, we will either ask questions
        or spawn child nodes for each subdirective.
        '''
        if completion.questions and len(completion.questions) > 0:
            message = f"An agent was tasked with decomposing the directive: \n`{self.node.message}`\n\nBefore decomposing, the agent decided it needs the following questions answered first:\n"
            message += "\n".join(completion.questions)

            self.add_journal_entry(
                {
                    "header": "Question Requested",
                    "content": message
                }
            )
            
            spawn_operation = SpawnOperation(
                node_id=self.node.node_id,
                node_embryo=NodeEmbryo(
                    action_id="swarmstar/actions/communication/ask_user_questions",
                    message=message,
                ),
            )
            return spawn_operation
        else:
            subdirectives = completion.subdirectives
            spawn_operations = []
            for subdirective in subdirectives:
                spawn_operation = SpawnOperation(
                    node_id=self.node.node_id,
                    node_embryo=NodeEmbryo(
                        action_id="swarmstar/actions/reasoning/route_action",
                        message=subdirective,
                    ),
                )
                spawn_operations.append(spawn_operation)
            subdirectives_str = "\n".join(subdirectives)
            self.add_journal_entry(
                {
                    "header": "Successfully Decomposed Directives",
                    "content": subdirectives_str,
                }
            )

            return spawn_operations



"""
The swarm consists of nodes. Each node is given a message 
and a preassigned action they must execute.
"""

from typing import Any, Dict, List, Optional

from pydantic import BaseModel
from typing_extensions import Literal


class SwarmNode(BaseModel):
    node_id: str
    parent_id: Optional[str] = None
    children_ids: List[str] = []
    action_id: str
    message: str
    alive: bool
    termination_policy: Literal[
        "simple",
        "parallel_review",
        "clone_with_questions_answered",
    ]
    developer_logs: List[Dict[str, Any]] = []
    journal: List[Dict[str, Any]] = []
    report: Optional[str] = None



from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def delete_file(swarm: SwarmConfig, file_path: str) -> None:
    try:
        os.remove(file_path)
    except Exception as e:
        raise ValueError(f"Failed to delete file: {str(e)}")


def move_file(swarm: SwarmConfig, file_path: str, new_file_path: str) -> None:
    try:
        shutil.move(file_path, new_file_path)
    except Exception as e:
        raise ValueError(f"Failed to move file: {str(e)}")


def rename_file(swarm: SwarmConfig, file_path: str, new_file_name: str) -> None:
    try:
        directory = os.path.dirname(file_path)
        new_file_path = os.path.join(directory, new_file_name)
        os.rename(file_path, new_file_path)
    except Exception as e:
        raise ValueError(f"Failed to rename file: {str(e)}")


def upload_file(swarm: SwarmConfig, file_path: str, file_bytes: bytes) -> None:
    try:
        with open(file_path, "wb") as file:
            file.write(file_bytes)
    except Exception as e:
        raise ValueError(f"Failed to upload file: {str(e)}")


def retrieve_file(swarm: SwarmConfig, file_path: str) -> bytes:
    try:
        with open(file_path, "rb") as file:
            file_bytes = file.read()
        return file_bytes
    except Exception as e:
        raise ValueError(f"Failed to retrieve file: {str(e)}")
